@@ -3,35 +3,39 @@ package main
 import (
 	"context"
 	"fmt"
-	"io/ioutil"
 	"net/http"
 	_ "net/http/pprof"
 	"os"
 	"os/signal"
 	"path/filepath"
-	gruntime "runtime"
+	"runtime"
+	"runtime/pprof"
 	"sort"
 	"strings"
 	"time"
 
-	_ "github.com/containers/libpod/v2/pkg/hooks/0.1.0"
+	_ "github.com/containers/podman/v3/pkg/hooks/0.1.0"
 	"github.com/containers/storage/pkg/reexec"
 	"github.com/cri-o/cri-o/internal/criocli"
 	"github.com/cri-o/cri-o/internal/log"
+	"github.com/cri-o/cri-o/internal/opentelemetry"
 	"github.com/cri-o/cri-o/internal/signals"
 	"github.com/cri-o/cri-o/internal/version"
 	libconfig "github.com/cri-o/cri-o/pkg/config"
 	"github.com/cri-o/cri-o/server"
+	v1 "github.com/cri-o/cri-o/server/cri/v1"
+	"github.com/cri-o/cri-o/server/cri/v1alpha2"
 	"github.com/cri-o/cri-o/server/metrics"
 	"github.com/cri-o/cri-o/utils"
 	grpc_middleware "github.com/grpc-ecosystem/go-grpc-middleware"
+	"github.com/pkg/errors"
 	"github.com/sirupsen/logrus"
 	"github.com/soheilhy/cmux"
 	"github.com/urfave/cli/v2"
+	"go.opentelemetry.io/contrib/instrumentation/google.golang.org/grpc/otelgrpc"
+	sdktrace "go.opentelemetry.io/otel/sdk/trace"
 	"golang.org/x/sys/unix"
 	"google.golang.org/grpc"
-	runtime "k8s.io/cri-api/pkg/apis/runtime/v1alpha2"
-	"k8s.io/klog/v2"
 )
 
 func writeCrioGoroutineStacks() {
@@ -44,7 +48,7 @@ func writeCrioGoroutineStacks() {
 	}
 }
 
-func catchShutdown(ctx context.Context, cancel context.CancelFunc, gserver *grpc.Server, sserver *server.Server, hserver *http.Server, signalled *bool) {
+func catchShutdown(ctx context.Context, cancel context.CancelFunc, gserver *grpc.Server, tp *sdktrace.TracerProvider, sserver *server.Server, hserver *http.Server, signalled *bool) {
 	sig := make(chan os.Signal, 2048)
 	signal.Notify(sig, signals.Interrupt, signals.Term, unix.SIGUSR1, unix.SIGUSR2, unix.SIGPIPE, signals.Hup)
 	go func() {
@@ -57,7 +61,7 @@ func catchShutdown(ctx context.Context, cancel context.CancelFunc, gserver *grpc
 				writeCrioGoroutineStacks()
 				continue
 			case unix.SIGUSR2:
-				gruntime.GC()
+				runtime.GC()
 				continue
 			case unix.SIGPIPE:
 				continue
@@ -69,15 +73,20 @@ func catchShutdown(ctx context.Context, cancel context.CancelFunc, gserver *grpc
 				continue
 			}
 			*signalled = true
+			if tp != nil {
+				if err := tp.Shutdown(ctx); err != nil {
+					logrus.Warnf("Error shutting down opentelemetry tracer provider: %v", err)
+				}
+			}
 			gserver.GracefulStop()
 			hserver.Shutdown(ctx) // nolint: errcheck
 			if err := sserver.StopStreamServer(); err != nil {
-				logrus.Warnf("error shutting down streaming server: %v", err)
+				logrus.Warnf("Error shutting down streaming server: %v", err)
 			}
 			sserver.StopMonitors()
 			cancel()
 			if err := sserver.Shutdown(ctx); err != nil {
-				logrus.Warnf("error shutting down main service %v", err)
+				logrus.Warnf("Error shutting down main service %v", err)
 			}
 			return
 		}
@@ -99,13 +108,10 @@ scope of the CRI.
 6. Resource isolation as required by the CRI.`
 
 func main() {
-	// Configure klog to not write any output
-	klog.LogToStderr(true)
-	klog.SetOutput(ioutil.Discard)
+	log.InitKlogShim()
 
 	if reexec.Init() {
-		fmt.Fprintf(os.Stderr, "unable to initialize container storage\n")
-		os.Exit(-1)
+		return
 	}
 	app := cli.NewApp()
 
@@ -129,6 +135,7 @@ func main() {
 	app.Commands = criocli.DefaultCommands
 	app.Commands = append(app.Commands, []*cli.Command{
 		configCommand,
+		publishCommand,
 		versionCommand,
 		wipeCommand,
 	}...)
@@ -139,12 +146,11 @@ func main() {
 			return err
 		}
 
-		cf := &logrus.TextFormatter{
+		logrus.SetFormatter(&logrus.TextFormatter{
 			TimestampFormat: "2006-01-02 15:04:05.000000000Z07:00",
 			FullTimestamp:   true,
-		}
-
-		logrus.SetFormatter(cf)
+		})
+		version.LogVersion()
 
 		level, err := logrus.ParseLevel(config.LogLevel)
 		if err != nil {
@@ -181,13 +187,32 @@ func main() {
 
 	app.Action = func(c *cli.Context) error {
 		ctx, cancel := context.WithCancel(context.Background())
+
+		cpuProfilePath := c.String("profile-cpu")
+		if cpuProfilePath != "" {
+			logrus.Infof("Creating CPU profile in: %v", cpuProfilePath)
+
+			file, err := os.Create(cpuProfilePath)
+			if err != nil {
+				cancel()
+				return errors.Wrap(err, "could not create CPU profile")
+			}
+			defer file.Close()
+
+			if err := pprof.StartCPUProfile(file); err != nil {
+				cancel()
+				return errors.Wrap(err, "could not start CPU profiling")
+			}
+			defer pprof.StopCPUProfile()
+		}
+
 		if c.Bool("profile") {
 			profilePort := c.Int("profile-port")
 			profileEndpoint := fmt.Sprintf("localhost:%v", profilePort)
 			go func() {
-				logrus.Debugf("starting profiling server on %v", profileEndpoint)
+				logrus.Debugf("Starting profiling server on %v", profileEndpoint)
 				if err := http.ListenAndServe(profileEndpoint, nil); err != nil {
-					logrus.Fatalf("unable to run profiling server: %v", err)
+					logrus.Fatalf("Unable to run profiling server: %v", err)
 				}
 			}()
 		}
@@ -211,20 +236,41 @@ func main() {
 
 		lis, err := server.Listen("unix", config.Listen)
 		if err != nil {
-			logrus.Fatalf("failed to listen: %v", err)
+			logrus.Fatalf("Failed to listen: %v", err)
 		}
 
-		grpcServer := grpc.NewServer(
-			grpc.UnaryInterceptor(grpc_middleware.ChainUnaryServer(
+		if err := os.Chmod(config.Listen, 0o660); err != nil {
+			logrus.Fatalf("Failed to chmod listen socket %s: %v", config.Listen, err)
+		}
+
+		var tracerProvider *sdktrace.TracerProvider
+		chainUnaryServer := grpc_middleware.ChainUnaryServer(metrics.UnaryInterceptor(), log.UnaryInterceptor())
+		chainStreamServer := grpc_middleware.ChainStreamServer(log.StreamInterceptor())
+		if config.EnableTracing {
+			var opts []otelgrpc.Option
+			tracerProvider, opts, err = opentelemetry.InitTracing(
+				ctx,
+				config.TracingEndpoint,
+				config.TracingSamplingRatePerMillion,
+			)
+			if err != nil {
+				logrus.Fatalf("Failed to initialize tracer provider: %v", err)
+			}
+			chainUnaryServer = grpc_middleware.ChainUnaryServer(
 				metrics.UnaryInterceptor(),
 				log.UnaryInterceptor(),
-			)),
-			grpc.StreamInterceptor(log.StreamInterceptor()),
+				otelgrpc.UnaryServerInterceptor(opts...),
+			)
+			chainStreamServer = grpc_middleware.ChainStreamServer(log.StreamInterceptor(), otelgrpc.StreamServerInterceptor(opts...))
+		}
+		grpcServer := grpc.NewServer(
+			grpc.UnaryInterceptor(chainUnaryServer),
+			grpc.StreamInterceptor(chainStreamServer),
 			grpc.MaxSendMsgSize(config.GRPCMaxSendMsgSize),
 			grpc.MaxRecvMsgSize(config.GRPCMaxRecvMsgSize),
 		)
 
-		service, err := server.New(ctx, config)
+		crioServer, err := server.New(ctx, config)
 		if err != nil {
 			logrus.Fatal(err)
 		}
@@ -241,20 +287,42 @@ func main() {
 			logrus.Fatal(err)
 		}
 
-		runtime.RegisterRuntimeServiceServer(grpcServer, service)
-		runtime.RegisterImageServiceServer(grpcServer, service)
+		if config.CleanShutdownFile != "" {
+			// clear out the shutdown file
+			if err := os.Remove(config.CleanShutdownFile); err != nil && !os.IsNotExist(err) {
+				logrus.Error(err)
+			}
+
+			// Write "$CleanShutdownFile".supported to show crio-wipe that
+			// we should be wiping if the CleanShutdownFile wasn't found.
+			// This protects us from wiping after an upgrade from a version that don't support
+			// CleanShutdownFile.
+			f, err := os.Create(config.CleanShutdownSupportedFileName())
+			if err != nil {
+				logrus.Errorf("Writing clean shutdown supported file: %v", err)
+			}
+			f.Close()
+
+			// and sync the changes to disk
+			if err := utils.SyncParent(config.CleanShutdownFile); err != nil {
+				logrus.Errorf("Failed to sync parent directory of clean shutdown file: %v", err)
+			}
+		}
+
+		v1alpha2.Register(grpcServer, crioServer)
+		v1.Register(grpcServer, crioServer)
 
 		// after the daemon is done setting up we can notify systemd api
 		notifySystem()
 
 		go func() {
-			service.StartExitMonitor()
+			crioServer.StartExitMonitor(ctx)
 		}()
 		hookSync := make(chan error, 2)
-		if service.ContainerServer.Hooks == nil {
+		if crioServer.ContainerServer.Hooks == nil {
 			hookSync <- err // so we don't block during cleanup
 		} else {
-			go service.ContainerServer.Hooks.Monitor(ctx, hookSync)
+			go crioServer.ContainerServer.Hooks.Monitor(ctx, hookSync)
 			err = <-hookSync
 			if err != nil {
 				cancel()
@@ -266,23 +334,23 @@ func main() {
 		grpcL := m.MatchWithWriters(cmux.HTTP2MatchHeaderFieldSendSettings("content-type", "application/grpc"))
 		httpL := m.Match(cmux.HTTP1Fast())
 
-		infoMux := service.GetInfoMux(c.Bool("enable-profile-unix-socket"))
+		infoMux := crioServer.GetExtendInterfaceMux(c.Bool("enable-profile-unix-socket"))
 		httpServer := &http.Server{
 			Handler:     infoMux,
 			ReadTimeout: 5 * time.Second,
 		}
 
 		graceful := false
-		catchShutdown(ctx, cancel, grpcServer, service, httpServer, &graceful)
+		catchShutdown(ctx, cancel, grpcServer, tracerProvider, crioServer, httpServer, &graceful)
 
 		go func() {
 			if err := grpcServer.Serve(grpcL); err != nil {
-				logrus.Errorf("unable to run GRPC server: %v", err)
+				logrus.Errorf("Unable to run GRPC server: %v", err)
 			}
 		}()
 		go func() {
 			if err := httpServer.Serve(httpL); err != nil {
-				logrus.Debugf("closed http server")
+				logrus.Debugf("Closed http server")
 			}
 		}()
 
@@ -298,38 +366,51 @@ func main() {
 			}
 		}()
 
-		streamServerCloseCh := service.StreamingServerCloseChan()
-		serverMonitorsCh := service.MonitorsCloseChan()
+		streamServerCloseCh := crioServer.StreamingServerCloseChan()
+		serverMonitorsCh := crioServer.MonitorsCloseChan()
 		select {
 		case <-streamServerCloseCh:
 		case <-serverMonitorsCh:
 		case <-serverCloseCh:
 		}
 
-		if err := service.Shutdown(ctx); err != nil {
-			logrus.Warnf("error shutting down service: %v", err)
+		if err := crioServer.Shutdown(ctx); err != nil {
+			logrus.Warnf("Error shutting down service: %v", err)
 		}
 		cancel()
 
 		<-streamServerCloseCh
-		logrus.Debug("closed stream server")
+		logrus.Debugf("Closed stream server")
 		<-serverMonitorsCh
-		logrus.Debug("closed monitors")
+		logrus.Debugf("Closed monitors")
 		err = <-hookSync
 		if err == nil || err == context.Canceled {
-			logrus.Debug("closed hook monitor")
+			logrus.Debugf("Closed hook monitor")
 		} else {
-			logrus.Errorf("hook monitor failed: %v", err)
+			logrus.Errorf("Hook monitor failed: %v", err)
 		}
 		<-serverCloseCh
-		logrus.Debug("closed main server")
+		logrus.Debugf("Closed main server")
+
+		memProfilePath := c.String("profile-mem")
+		if memProfilePath != "" {
+			logrus.Infof("Creating memory profile in: %v", memProfilePath)
+
+			file, err := os.Create(memProfilePath)
+			if err != nil {
+				return errors.Wrap(err, "could not create memory profile")
+			}
+			defer file.Close()
+			runtime.GC()
+
+			if err := pprof.WriteHeapProfile(file); err != nil {
+				return errors.Wrap(err, "could not write memory profile")
+			}
+		}
 
 		return nil
 	}
 
-	// Log our version early at startup
-	version.LogVersion()
-
 	if err := app.Run(os.Args); err != nil {
 		logrus.Fatal(err)
 	}
