@@ -1,4 +1,4 @@
 #!/usr/bin/env bash
-if ! gpgme-config --libs &>/dev/null; then
+if gpgme-config --libs &>/dev/null; then
     echo containers_image_openpgp
 fi
