@@ -1,3 +1,4 @@
+//go:build linux
 // +build linux
 
 package cgmgr
@@ -10,9 +11,13 @@ import (
 	"strconv"
 	"strings"
 
+	"github.com/containers/podman/v3/pkg/cgroups"
 	"github.com/cri-o/cri-o/internal/config/node"
+	libctr "github.com/opencontainers/runc/libcontainer/cgroups"
+	rspec "github.com/opencontainers/runtime-spec/specs-go"
 	"github.com/pkg/errors"
 	"github.com/sirupsen/logrus"
+	types "k8s.io/cri-api/pkg/apis/runtime/v1"
 )
 
 const (
@@ -52,15 +57,24 @@ type CgroupManager interface {
 	// returns the cgroup path on disk for that containerID. If parentCgroup is empty, it
 	// uses the default parent for that particular manager
 	ContainerCgroupAbsolutePath(string, string) (string, error)
+	// PopulateContainerCgroupStats fills the stats object with information from the cgroup found
+	// given a cgroup parent and container ID.
+	PopulateContainerCgroupStats(sbParent, containerID string, stats *types.ContainerStats) error
 	// SandboxCgroupPath takes the sandbox parent, and sandbox ID. It
 	// returns the cgroup parent, cgroup path, and error. For systemd cgroups,
 	// it also checks there is enough memory in the given cgroup
 	SandboxCgroupPath(string, string) (string, string, error)
-	// MoveConmonToCgroup takes the container ID, cgroup parent, conmon's cgroup (from the config) and conmon's PID
-	// It attempts to move conmon to the correct cgroup.
+	// PopulateContainerCgroupStats takes arguments sandbox parent cgroup, and sandbox stats object.
+	// It fills the object with information from the cgroup found given that parent.
+	PopulateSandboxCgroupStats(sbParent string, stats *types.PodSandboxStats) error
+	// MoveConmonToCgroup takes the container ID, cgroup parent, conmon's cgroup (from the config), conmon's PID, and some customized resources
+	// It attempts to move conmon to the correct cgroup, and set the resources for that cgroup.
 	// It returns the cgroupfs parent that conmon was put into
 	// so that CRI-O can clean the parent cgroup of the newly added conmon once the process terminates (systemd handles this for us)
-	MoveConmonToCgroup(cid, cgroupParent, conmonCgroup string, pid int) (string, error)
+	MoveConmonToCgroup(cid, cgroupParent, conmonCgroup string, pid int, resources *rspec.LinuxResources) (string, error)
+	// CreateSandboxCgroup takes the sandbox parent, and sandbox ID.
+	// It creates a new cgroup for that sandbox, which is useful when spoofing an infra container.
+	CreateSandboxCgroup(sbParent, containerID string) error
 }
 
 // New creates a new CgroupManager with defaults
@@ -126,3 +140,36 @@ func VerifyMemoryIsEnough(memoryLimit int64) error {
 	}
 	return nil
 }
+
+// createSandboxCgroup takes the sandbox parent, and sandbox ID.
+// It creates a new cgroup for that sandbox, which is useful when spoofing an infra container.
+func createSandboxCgroup(sbParent, containerID string, mgr CgroupManager) error {
+	path, err := mgr.ContainerCgroupAbsolutePath(sbParent, containerID)
+	if err != nil {
+		return err
+	}
+	_, err = cgroups.New(path, &rspec.LinuxResources{})
+	return err
+}
+
+// MoveProcessToContainerCgroup moves process to the container cgroup
+func MoveProcessToContainerCgroup(containerPid, commandPid int) error {
+	parentCgroupFile := fmt.Sprintf("/proc/%d/cgroup", containerPid)
+	cgmap, err := libctr.ParseCgroupFile(parentCgroupFile)
+	if err != nil {
+		return err
+	}
+
+	var dir string
+	for controller, path := range cgmap {
+		// For cgroups V2, controller will be an empty string
+		dir = filepath.Join("/sys/fs/cgroup", controller, path)
+
+		if libctr.PathExists(dir) {
+			if err := libctr.WriteCgroupProc(dir, commandPid); err != nil {
+				return err
+			}
+		}
+	}
+	return nil
+}
