@@ -12,10 +12,10 @@ import (
 	"time"
 
 	"github.com/cri-o/cri-o/pkg/config"
-	"github.com/cri-o/cri-o/server/cri/types"
 	rspec "github.com/opencontainers/runtime-spec/specs-go"
 	"github.com/pkg/errors"
 	"golang.org/x/net/context"
+	types "k8s.io/cri-api/pkg/apis/runtime/v1"
 
 	"k8s.io/client-go/tools/remotecommand"
 )
@@ -35,6 +35,11 @@ const (
 	// killContainerTimeout is the timeout that we wait for the container to
 	// be SIGKILLed.
 	killContainerTimeout = 2 * time.Minute
+
+	// maxExecSyncSize is the maximum size of exec sync output CRI-O will process.
+	// It is set to the amount of logs allowed in the dockershim implementation:
+	// https://github.com/kubernetes/kubernetes/pull/82514
+	maxExecSyncSize = 16 * 1024 * 1024
 )
 
 // Runtime is the generic structure holding both global and specific
@@ -64,14 +69,13 @@ type RuntimeImpl interface {
 	UpdateContainerStatus(context.Context, *Container) error
 	PauseContainer(context.Context, *Container) error
 	UnpauseContainer(context.Context, *Container) error
-	ContainerStats(context.Context, *Container, string) (*ContainerStats, error)
+	ContainerStats(context.Context, *Container, string) (*types.ContainerStats, error)
 	SignalContainer(context.Context, *Container, syscall.Signal) error
 	AttachContainer(context.Context, *Container, io.Reader, io.WriteCloser, io.WriteCloser,
 		bool, <-chan remotecommand.TerminalSize) error
 	PortForwardContainer(context.Context, *Container, string,
 		int32, io.ReadWriteCloser) error
 	ReopenContainerLog(context.Context, *Container) error
-	WaitContainerStateStopped(context.Context, *Container) error
 }
 
 // New creates a new Runtime with options provided
@@ -111,64 +115,6 @@ func (r *Runtime) ValidateRuntimeHandler(handler string) (*config.RuntimeHandler
 	return runtimeHandler, nil
 }
 
-// WaitContainerStateStopped runs a loop polling UpdateStatus(), seeking for
-// the container status to be updated to 'stopped'. Either it gets the expected
-// status and returns nil, or it reaches the timeout and returns an error.
-func (r *Runtime) WaitContainerStateStopped(ctx context.Context, c *Container) error {
-	impl, err := r.RuntimeImpl(c)
-	if err != nil {
-		return err
-	}
-
-	// No need to go further and spawn the go routine if the container
-	// is already in the expected status.
-	if c.State().Status == ContainerStateStopped {
-		return nil
-	}
-
-	done := make(chan error)
-	chControl := make(chan struct{})
-	go func() {
-		defer close(done)
-		for {
-			select {
-			case <-chControl:
-				return
-			default:
-				// Check if the container is stopped
-				if err := impl.UpdateContainerStatus(ctx, c); err != nil {
-					done <- err
-					return
-				}
-				if c.State().Status == ContainerStateStopped {
-					done <- nil
-					return
-				}
-				time.Sleep(100 * time.Millisecond)
-			}
-		}
-	}()
-	select {
-	case err = <-done:
-		break
-	case <-ctx.Done():
-		close(chControl)
-		return ctx.Err()
-	case <-time.After(time.Duration(r.config.CtrStopTimeout) * time.Second):
-		close(chControl)
-		return fmt.Errorf(
-			"failed to get container stopped status: %ds timeout reached",
-			r.config.CtrStopTimeout,
-		)
-	}
-
-	if err != nil {
-		return fmt.Errorf("failed to get container stopped status: %v", err)
-	}
-
-	return nil
-}
-
 func (r *Runtime) getRuntimeHandler(handler string) (*config.RuntimeHandler, error) {
 	// Define the current runtime handler as the default runtime handler.
 	rh := r.config.Runtimes[r.config.DefaultRuntime]
@@ -200,23 +146,14 @@ func (r *Runtime) PrivilegedWithoutHostDevices(handler string) (bool, error) {
 	return rh.PrivilegedWithoutHostDevices, nil
 }
 
-// FilterDisallowedAnnotations filters annotations that are not specified in the allowed_annotations map
-// for a given handler.
-// This function returns an error if the runtime handler can't be found.
-// The annotations map is mutated in-place.
-func (r *Runtime) FilterDisallowedAnnotations(handler string, annotations map[string]string) error {
+// AllowedAnnotations returns the allowed annotations for this runtime.
+func (r *Runtime) AllowedAnnotations(handler string) ([]string, error) {
 	rh, err := r.getRuntimeHandler(handler)
 	if err != nil {
-		return err
+		return []string{}, err
 	}
-	for ann := range annotations {
-		for _, disallowed := range rh.DisallowedAnnotations {
-			if strings.HasPrefix(ann, disallowed) {
-				delete(annotations, disallowed)
-			}
-		}
-	}
-	return nil
+
+	return rh.AllowedAnnotations, nil
 }
 
 // RuntimeType returns the type of runtimeHandler
@@ -241,6 +178,10 @@ func (r *Runtime) newRuntimeImpl(c *Container) (RuntimeImpl, error) {
 		return newRuntimeVM(rh.RuntimePath, rh.RuntimeRoot, rh.RuntimeConfigPath), nil
 	}
 
+	if rh.RuntimeType == config.RuntimeTypePod {
+		return newRuntimePod(r, rh, c)
+	}
+
 	// If the runtime type is different from "vm", then let's fallback
 	// onto the OCI implementation by default.
 	return newRuntimeOCI(r, rh), nil
@@ -251,10 +192,17 @@ func (r *Runtime) RuntimeImpl(c *Container) (RuntimeImpl, error) {
 	r.runtimeImplMapMutex.RLock()
 	impl, ok := r.runtimeImplMap[c.ID()]
 	r.runtimeImplMapMutex.RUnlock()
-	if !ok {
-		return r.newRuntimeImpl(c)
+	if ok {
+		return impl, nil
 	}
 
+	r.runtimeImplMapMutex.Lock()
+	defer r.runtimeImplMapMutex.Unlock()
+	impl, err := r.newRuntimeImpl(c)
+	if err != nil {
+		return nil, err
+	}
+	r.runtimeImplMap[c.ID()] = impl
 	return impl, nil
 }
 
@@ -377,7 +325,7 @@ func (r *Runtime) UnpauseContainer(ctx context.Context, c *Container) error {
 }
 
 // ContainerStats provides statistics of a container.
-func (r *Runtime) ContainerStats(ctx context.Context, c *Container, cgroup string) (*ContainerStats, error) {
+func (r *Runtime) ContainerStats(ctx context.Context, c *Container, cgroup string) (*types.ContainerStats, error) {
 	impl, err := r.RuntimeImpl(c)
 	if err != nil {
 		return nil, err
