@@ -1,6 +1,3 @@
-//go:build linux
-// +build linux
-
 package server
 
 import (
@@ -9,7 +6,6 @@ import (
 	"path/filepath"
 	"sort"
 	"strings"
-	"syscall"
 	"time"
 
 	"github.com/containers/buildah/util"
@@ -23,19 +19,20 @@ import (
 	"github.com/cri-o/cri-o/internal/config/device"
 	"github.com/cri-o/cri-o/internal/config/node"
 	"github.com/cri-o/cri-o/internal/config/rdt"
+	ctrfactory "github.com/cri-o/cri-o/internal/factory/container"
 	"github.com/cri-o/cri-o/internal/lib/sandbox"
 	"github.com/cri-o/cri-o/internal/log"
 	oci "github.com/cri-o/cri-o/internal/oci"
 	"github.com/cri-o/cri-o/internal/storage"
 	crioann "github.com/cri-o/cri-o/pkg/annotations"
-	ctrIface "github.com/cri-o/cri-o/pkg/container"
-	"github.com/cri-o/cri-o/server/cri/types"
 	securejoin "github.com/cyphar/filepath-securejoin"
 	rspec "github.com/opencontainers/runtime-spec/specs-go"
 	"github.com/opencontainers/runtime-tools/generate"
 	"github.com/pkg/errors"
 	"github.com/sirupsen/logrus"
 	"golang.org/x/net/context"
+	"golang.org/x/sys/unix"
+	types "k8s.io/cri-api/pkg/apis/runtime/v1"
 
 	"github.com/intel/goresctrl/pkg/blockio"
 )
@@ -69,18 +66,20 @@ func makeAccessible(path string, uid, gid int, doChown bool) error {
 		}
 	}
 	for ; path != "/"; path = filepath.Dir(path) {
-		st, err := os.Stat(path)
+		var st unix.Stat_t
+		err := unix.Stat(path, &st)
 		if err != nil {
 			if os.IsNotExist(err) {
 				return nil
 			}
 			return err
 		}
-		if int(st.Sys().(*syscall.Stat_t).Uid) == uid && int(st.Sys().(*syscall.Stat_t).Gid) == gid {
+		if int(st.Uid) == uid && int(st.Gid) == gid {
 			continue
 		}
-		if st.Mode()&0o111 != 0o111 {
-			if err := os.Chmod(path, st.Mode()|0o111); err != nil {
+		perm := os.FileMode(st.Mode) & os.ModePerm
+		if perm&0o111 != 0o111 {
+			if err := os.Chmod(path, perm|0o111); err != nil {
 				return err
 			}
 		}
@@ -136,7 +135,7 @@ func (s *Server) finalizeUserMapping(specgen *generate.Generator, mappings *idto
 	}
 }
 
-func (s *Server) createSandboxContainer(ctx context.Context, ctr ctrIface.Container, sb *sandbox.Sandbox) (cntr *oci.Container, retErr error) {
+func (s *Server) createSandboxContainer(ctx context.Context, ctr ctrfactory.Container, sb *sandbox.Sandbox) (cntr *oci.Container, retErr error) {
 	// TODO: simplify this function (cyclomatic complexity here is high)
 	// TODO: factor generating/updating the spec into something other projects can vendor
 
@@ -146,7 +145,7 @@ func (s *Server) createSandboxContainer(ctx context.Context, ctr ctrIface.Contai
 	// TODO: eventually, this should be in the container package, but it's going through a lot of churn
 	// and SpecAddAnnotations is already being passed too many arguments
 	// Filter early so any use of the annotations don't use the wrong values
-	if err := s.Runtime().FilterDisallowedAnnotations(sb.RuntimeHandler(), ctr.Config().Annotations); err != nil {
+	if err := s.FilterDisallowedAnnotations(sb.Annotations(), ctr.Config().Annotations, sb.RuntimeHandler()); err != nil {
 		return nil, err
 	}
 
@@ -156,6 +155,12 @@ func (s *Server) createSandboxContainer(ctx context.Context, ctr ctrIface.Contai
 	if err := ctr.SetPrivileged(); err != nil {
 		return nil, err
 	}
+	if containerConfig.Linux == nil {
+		containerConfig.Linux = &types.LinuxContainerConfig{}
+	}
+	if containerConfig.Linux.SecurityContext == nil {
+		containerConfig.Linux.SecurityContext = newLinuxContainerSecurityContext()
+	}
 	securityContext := containerConfig.Linux.SecurityContext
 
 	// creates a spec Generator with the default spec.
@@ -270,9 +275,12 @@ func (s *Server) createSandboxContainer(ctx context.Context, ctr ctrIface.Contai
 	if !ctr.Privileged() {
 		processLabel = containerInfo.ProcessLabel
 	}
-	hostIPC := securityContext.NamespaceOptions.Ipc == types.NamespaceModeNODE
-	hostPID := securityContext.NamespaceOptions.Pid == types.NamespaceModeNODE
-	hostNet := securityContext.NamespaceOptions.Network == types.NamespaceModeNODE
+	if securityContext.NamespaceOptions == nil {
+		securityContext.NamespaceOptions = &types.NamespaceOption{}
+	}
+	hostIPC := securityContext.NamespaceOptions.Ipc == types.NamespaceMode_NODE
+	hostPID := securityContext.NamespaceOptions.Pid == types.NamespaceMode_NODE
+	hostNet := securityContext.NamespaceOptions.Network == types.NamespaceMode_NODE
 
 	// Don't use SELinux separation with Host Pid or IPC Namespace or privileged.
 	if hostPID || hostIPC {
@@ -290,13 +298,21 @@ func (s *Server) createSandboxContainer(ctx context.Context, ctr ctrIface.Contai
 
 	skipRelabel := false
 	const superPrivilegedType = "spc_t"
+	if securityContext.SelinuxOptions == nil {
+		securityContext.SelinuxOptions = &types.SELinuxOption{}
+	}
 	if securityContext.SelinuxOptions.Type == superPrivilegedType || // super privileged container
-		(ctr.SandboxConfig().Linux.SecurityContext.SelinuxOptions.Type == superPrivilegedType && // super privileged pod
+		(ctr.SandboxConfig().Linux != nil &&
+			ctr.SandboxConfig().Linux.SecurityContext != nil &&
+			ctr.SandboxConfig().Linux.SecurityContext.SelinuxOptions != nil &&
+			ctr.SandboxConfig().Linux.SecurityContext.SelinuxOptions.Type == superPrivilegedType && // super privileged pod
 			securityContext.SelinuxOptions.Type == "") {
 		skipRelabel = true
 	}
 
-	containerVolumes, ociMounts, err := addOCIBindMounts(ctx, ctr, mountLabel, s.config.RuntimeConfig.BindMountPrefix, s.config.AbsentMountSourcesToReject, maybeRelabel, skipRelabel)
+	cgroup2RW := node.CgroupIsV2() && sb.Annotations()[crioann.Cgroup2RWAnnotation] == "true"
+
+	containerVolumes, ociMounts, err := addOCIBindMounts(ctx, ctr, mountLabel, s.config.RuntimeConfig.BindMountPrefix, s.config.AbsentMountSourcesToReject, maybeRelabel, skipRelabel, cgroup2RW)
 	if err != nil {
 		return nil, err
 	}
@@ -308,7 +324,7 @@ func (s *Server) createSandboxContainer(ctx context.Context, ctr ctrIface.Contai
 		return nil, err
 	}
 
-	annotationDevices, err := device.DevicesFromAnnotation(sb.Annotations()[crioann.DevicesAnnotation])
+	annotationDevices, err := device.DevicesFromAnnotation(sb.Annotations()[crioann.DevicesAnnotation], s.config.AllowedDevices)
 	if err != nil {
 		return nil, err
 	}
@@ -362,9 +378,9 @@ func (s *Server) createSandboxContainer(ctx context.Context, ctr ctrIface.Contai
 	if linux != nil {
 		resources := linux.Resources
 		if resources != nil {
-			specgen.SetLinuxResourcesCPUPeriod(uint64(resources.CPUPeriod))
-			specgen.SetLinuxResourcesCPUQuota(resources.CPUQuota)
-			specgen.SetLinuxResourcesCPUShares(uint64(resources.CPUShares))
+			specgen.SetLinuxResourcesCPUPeriod(uint64(resources.CpuPeriod))
+			specgen.SetLinuxResourcesCPUQuota(resources.CpuQuota)
+			specgen.SetLinuxResourcesCPUShares(uint64(resources.CpuShares))
 
 			memoryLimit := resources.MemoryLimitInBytes
 			if memoryLimit != 0 {
@@ -372,14 +388,27 @@ func (s *Server) createSandboxContainer(ctx context.Context, ctr ctrIface.Contai
 					return nil, err
 				}
 				specgen.SetLinuxResourcesMemoryLimit(memoryLimit)
+				if resources.MemorySwapLimitInBytes != 0 {
+					if resources.MemorySwapLimitInBytes < resources.MemoryLimitInBytes {
+						return nil, errors.Errorf(
+							"container %s create failed because memory swap limit (%d) cannot be lower than memory limit (%d)",
+							ctr.ID(),
+							resources.MemorySwapLimitInBytes,
+							resources.MemoryLimitInBytes,
+						)
+					}
+					memoryLimit = resources.MemorySwapLimitInBytes
+				}
+				// If node doesn't have memory swap, then skip setting
+				// otherwise the container creation fails.
 				if node.CgroupHasMemorySwap() {
 					specgen.SetLinuxResourcesMemorySwap(memoryLimit)
 				}
 			}
 
 			specgen.SetProcessOOMScoreAdj(int(resources.OomScoreAdj))
-			specgen.SetLinuxResourcesCPUCpus(resources.CPUsetCPUs)
-			specgen.SetLinuxResourcesCPUMems(resources.CPUsetMems)
+			specgen.SetLinuxResourcesCPUCpus(resources.CpusetCpus)
+			specgen.SetLinuxResourcesCPUMems(resources.CpusetMems)
 
 			// If the kernel has no support for hugetlb, silently ignore the limits
 			if node.CgroupHasHugetlb() {
@@ -388,6 +417,15 @@ func (s *Server) createSandboxContainer(ctx context.Context, ctr ctrIface.Contai
 					specgen.AddLinuxResourcesHugepageLimit(limit.PageSize, limit.Limit)
 				}
 			}
+
+			if node.CgroupIsV2() && len(resources.Unified) != 0 {
+				if specgen.Config.Linux.Resources.Unified == nil {
+					specgen.Config.Linux.Resources.Unified = make(map[string]string, len(resources.Unified))
+				}
+				for key, value := range resources.Unified {
+					specgen.Config.Linux.Resources.Unified[key] = value
+				}
+			}
 		}
 
 		specgen.SetLinuxCgroupsPath(s.config.CgroupManager().ContainerCgroupPath(sb.CgroupParent(), containerID))
@@ -396,15 +434,7 @@ func (s *Server) createSandboxContainer(ctx context.Context, ctr ctrIface.Contai
 			specgen.SetupPrivileged(true)
 		} else {
 			capabilities := securityContext.Capabilities
-			// Ensure we don't get a nil pointer error if the config
-			// doesn't set any capabilities
-			if capabilities == nil {
-				capabilities = &types.Capability{}
-			}
-			// Clear default capabilities from spec
-			specgen.ClearProcessCapabilities()
-			err = setupCapabilities(specgen, capabilities, s.config.DefaultCapabilities)
-			if err != nil {
+			if err := ctr.SpecSetupCapabilities(capabilities, s.config.DefaultCapabilities); err != nil {
 				return nil, err
 			}
 		}
@@ -455,36 +485,26 @@ func (s *Server) createSandboxContainer(ctx context.Context, ctr ctrIface.Contai
 		return nil, err
 	}
 
-	// Join the namespace paths for the pod sandbox container.
-	if err := configureGeneratorGivenNamespacePaths(sb.NamespacePaths(), specgen); err != nil {
-		return nil, errors.Wrap(err, "failed to configure namespaces in container create")
+	var nsTargetCtr *oci.Container
+	if target := containerConfig.Linux.SecurityContext.NamespaceOptions.TargetId; target != "" {
+		nsTargetCtr = s.GetContainer(target)
 	}
 
-	if securityContext.NamespaceOptions.Pid == types.NamespaceModeNODE {
-		// kubernetes PodSpec specify to use Host PID namespace
-		if err := specgen.RemoveLinuxNamespace(string(rspec.PIDNamespace)); err != nil {
-			return nil, err
-		}
-	} else if securityContext.NamespaceOptions.Pid == types.NamespaceModePOD {
-		pidNsPath := sb.PidNsPath()
-		if pidNsPath == "" {
-			if sb.NamespaceOptions().Pid != types.NamespaceModePOD {
-				return nil, errors.New("Pod level PID namespace requested for the container, but pod sandbox was not similarly configured, and does not have an infra container")
+	if err := ctr.SpecAddNamespaces(sb, nsTargetCtr, &s.config); err != nil {
+		return nil, err
+	}
+
+	defer func() {
+		if retErr != nil && ctr.PidNamespace() != nil {
+			log.Infof(ctx, "CreateCtrLinux: clearing PID namespace for container %s", containerInfo.ID)
+			if err := ctr.PidNamespace().Remove(); err != nil {
+				log.Warnf(ctx, "Failed to remove PID namespace: %v", err)
 			}
-			return nil, errors.New("PID namespace requested, but sandbox infra container unexpectedly invalid")
 		}
-
-		if err := specgen.AddOrReplaceLinuxNamespace(string(rspec.PIDNamespace), pidNsPath); err != nil {
-			return nil, err
-		}
-	}
+	}()
 
 	// If the sandbox is configured to run in the host network, do not create a new network namespace
 	if hostNet {
-		if err := specgen.RemoveLinuxNamespace(string(rspec.NetworkNamespace)); err != nil {
-			return nil, err
-		}
-
 		if !isInCRIMounts("/sys", containerConfig.Mounts) {
 			ctr.SpecAddMount(rspec.Mount{
 				Destination: "/sys",
@@ -568,6 +588,18 @@ func (s *Server) createSandboxContainer(ctx context.Context, ctr ctrIface.Contai
 		})
 	}
 
+	if sb.ContainerEnvPath() != "" {
+		if err := securityLabel(sb.ContainerEnvPath(), mountLabel, false, false); err != nil {
+			return nil, err
+		}
+		ctr.SpecAddMount(rspec.Mount{
+			Destination: "/run/.containerenv",
+			Type:        "bind",
+			Source:      sb.ContainerEnvPath(),
+			Options:     append(options, "bind"),
+		})
+	}
+
 	if !isInCRIMounts("/etc/hosts", containerConfig.Mounts) && hostNet {
 		// Only bind mount for host netns and when CRI does not give us any hosts file
 		ctr.SpecAddMount(rspec.Mount{
@@ -677,7 +709,7 @@ func (s *Server) createSandboxContainer(ctx context.Context, ctr ctrIface.Contai
 		mountLabel,
 		containerInfo.RunDir,
 		s.config.DefaultMountsFile,
-		containerInfo.RunDir,
+		mountPoint,
 		0,
 		0,
 		rootless.IsRootless(),
@@ -733,11 +765,11 @@ func (s *Server) createSandboxContainer(ctx context.Context, ctr ctrIface.Contai
 
 	crioAnnotations := specgen.Config.Annotations
 
-	ociMetadata := &oci.Metadata{
+	criMetadata := &types.ContainerMetadata{
 		Name:    metadata.Name,
 		Attempt: metadata.Attempt,
 	}
-	ociContainer, err := oci.NewContainer(containerID, containerName, containerInfo.RunDir, logPath, labels, crioAnnotations, ctr.Config().Annotations, image, imageName, imageRef, ociMetadata, sb.ID(), containerConfig.Tty, containerConfig.Stdin, containerConfig.StdinOnce, sb.RuntimeHandler(), containerInfo.Dir, created, containerImageConfig.Config.StopSignal)
+	ociContainer, err := oci.NewContainer(containerID, containerName, containerInfo.RunDir, logPath, labels, crioAnnotations, ctr.Config().Annotations, image, imageName, imageRef, criMetadata, sb.ID(), containerConfig.Tty, containerConfig.Stdin, containerConfig.StdinOnce, sb.RuntimeHandler(), containerInfo.Dir, created, containerImageConfig.Config.StopSignal)
 	if err != nil {
 		return nil, err
 	}
@@ -745,6 +777,8 @@ func (s *Server) createSandboxContainer(ctx context.Context, ctr ctrIface.Contai
 	specgen.SetLinuxMountLabel(mountLabel)
 	specgen.SetProcessSelinuxLabel(processLabel)
 
+	ociContainer.AddManagedPIDNamespace(ctr.PidNamespace())
+
 	ociContainer.SetIDMappings(containerIDMappings)
 	var rootPair idtools.IDPair
 	if containerIDMappings != nil {
@@ -778,7 +812,7 @@ func (s *Server) createSandboxContainer(ctx context.Context, ctr ctrIface.Contai
 	// add symlink /etc/mtab to /proc/mounts allow looking for mountfiles there in the container
 	// compatible with Docker
 	mtab := filepath.Join(mountPoint, "/etc/mtab")
-	if err := idtools.MkdirAllAs(filepath.Dir(mtab), 0755, rootPair.UID, rootPair.GID); err != nil {
+	if err := idtools.MkdirAllAs(filepath.Dir(mtab), 0o755, rootPair.UID, rootPair.GID); err != nil {
 		return nil, errors.Wrap(err, "error creating mtab directory")
 	}
 	if err := os.Symlink("/proc/mounts", mtab); err != nil && !os.IsExist(err) {
@@ -848,7 +882,7 @@ func clearReadOnly(m *rspec.Mount) {
 	m.Options = append(m.Options, "rw")
 }
 
-func addOCIBindMounts(ctx context.Context, ctr ctrIface.Container, mountLabel, bindMountPrefix string, absentMountSourcesToReject []string, maybeRelabel, skipRelabel bool) ([]oci.ContainerVolume, []rspec.Mount, error) {
+func addOCIBindMounts(ctx context.Context, ctr ctrfactory.Container, mountLabel, bindMountPrefix string, absentMountSourcesToReject []string, maybeRelabel, skipRelabel, cgroup2RW bool) ([]oci.ContainerVolume, []rspec.Mount, error) {
 	volumes := []oci.ContainerVolume{}
 	ociMounts := []rspec.Mount{}
 	containerConfig := ctr.Config()
@@ -929,11 +963,11 @@ func addOCIBindMounts(ctx context.Context, ctr ctrIface.Container, mountLabel, b
 
 		// mount propagation
 		switch m.Propagation {
-		case types.MountPropagationPropagationPrivate:
+		case types.MountPropagation_PROPAGATION_PRIVATE:
 			options = append(options, "rprivate")
 			// Since default root propagation in runc is rprivate ignore
 			// setting the root propagation
-		case types.MountPropagationPropagationBidirectional:
+		case types.MountPropagation_PROPAGATION_BIDIRECTIONAL:
 			if err := ensureShared(src, mountInfos); err != nil {
 				return nil, nil, err
 			}
@@ -941,7 +975,7 @@ func addOCIBindMounts(ctx context.Context, ctr ctrIface.Container, mountLabel, b
 			if err := specgen.SetLinuxRootPropagation("rshared"); err != nil {
 				return nil, nil, err
 			}
-		case types.MountPropagationPropagationHostToContainer:
+		case types.MountPropagation_PROPAGATION_HOST_TO_CONTAINER:
 			if err := ensureSharedOrSlave(src, mountInfos); err != nil {
 				return nil, nil, err
 			}
@@ -983,7 +1017,13 @@ func addOCIBindMounts(ctx context.Context, ctr ctrIface.Container, mountLabel, b
 			Destination: cgroupSysFsPath,
 			Type:        "cgroup",
 			Source:      "cgroup",
-			Options:     []string{"nosuid", "noexec", "nodev", "relatime", "ro"},
+			Options:     []string{"nosuid", "noexec", "nodev", "relatime"},
+		}
+
+		if cgroup2RW {
+			m.Options = append(m.Options, "rw")
+		} else {
+			m.Options = append(m.Options, "ro")
 		}
 		specgen.AddMount(m)
 	}
@@ -1053,7 +1093,7 @@ func setupSystemd(mounts []rspec.Mount, g generate.Generator) {
 			}
 			g.AddMount(systemdMnt)
 		}
-		g.AddLinuxMaskedPaths("/sys/fs/cgroup/systemd/release_agent")
+		g.AddLinuxMaskedPaths(filepath.Join(cgroupSysFsSystemdPath, "release_agent"))
 	}
 	g.AddProcessEnv("container", "crio")
 }
@@ -1075,3 +1115,15 @@ func isBindMount(mountOptions []string) bool {
 	}
 	return false
 }
+
+func newLinuxContainerSecurityContext() *types.LinuxContainerSecurityContext {
+	return &types.LinuxContainerSecurityContext{
+		Capabilities:     &types.Capability{},
+		NamespaceOptions: &types.NamespaceOption{},
+		SelinuxOptions:   &types.SELinuxOption{},
+		RunAsUser:        &types.Int64Value{},
+		RunAsGroup:       &types.Int64Value{},
+		Seccomp:          &types.SecurityProfile{},
+		Apparmor:         &types.SecurityProfile{},
+	}
+}
