@@ -10,10 +10,11 @@ import (
 	imageTypes "github.com/containers/image/v5/types"
 	"github.com/cri-o/cri-o/internal/log"
 	"github.com/cri-o/cri-o/internal/storage"
-	"github.com/cri-o/cri-o/server/cri/types"
 	"github.com/cri-o/cri-o/server/metrics"
+	"github.com/cri-o/cri-o/utils"
 	"github.com/docker/distribution/registry/api/errcode"
 	"github.com/pkg/errors"
+	types "k8s.io/cri-api/pkg/apis/runtime/v1"
 )
 
 var localRegistryPrefix = "localhost/"
@@ -143,7 +144,7 @@ func (s *Server) pullImage(ctx context.Context, pullArgs *pullArguments) (string
 			tryIncrementImagePullFailureMetric(img, err)
 			continue
 		}
-		defer tmpImg.Close()
+		defer tmpImg.Close() // nolint:gocritic
 
 		var storedImage *storage.ImageResult
 		storedImage, err = s.StorageImageServer().ImageStatus(s.config.SystemContext, img)
@@ -163,6 +164,8 @@ func (s *Server) pullImage(ctx context.Context, pullArgs *pullArguments) (string
 				// Skipped bytes metrics
 				if storedImage.Size != nil {
 					metrics.Instance().MetricImagePullsByNameSkippedAdd(float64(*storedImage.Size), img)
+					// Metrics for image pull skipped bytes
+					metrics.Instance().MetricImagePullsSkippedBytesAdd(float64(*storedImage.Size))
 				}
 
 				break
@@ -172,7 +175,7 @@ func (s *Server) pullImage(ctx context.Context, pullArgs *pullArguments) (string
 
 		// Pull by collecting progress metrics
 		progress := make(chan imageTypes.ProgressProperties)
-		defer close(progress)
+		defer close(progress) // nolint:gocritic
 		go func() {
 			for p := range progress {
 				if p.Event == imageTypes.ProgressEventSkipped {
@@ -203,6 +206,13 @@ func (s *Server) pullImage(ctx context.Context, pullArgs *pullArguments) (string
 					img, fmt.Sprintf("%d", imageSize(tmpImg)),
 				)
 
+				// Metrics for image pulls bytes
+				metrics.Instance().MetricImagePullsBytesAdd(
+					float64(p.OffsetUpdate),
+					p.Artifact.MediaType,
+					p.Artifact.Size,
+				)
+
 				// Metrics for size histogram
 				if p.Event == imageTypes.ProgressEventDone {
 					metrics.Instance().MetricImagePullsLayerSizeObserve(p.Artifact.Size)
@@ -216,7 +226,7 @@ func (s *Server) pullImage(ctx context.Context, pullArgs *pullArguments) (string
 			if !s.config.CgroupManager().IsSystemd() {
 				return "", errors.New("--separate-pull-cgroup is supported only with systemd")
 			}
-			if s.config.SeparatePullCgroup == "pod" {
+			if s.config.SeparatePullCgroup == utils.PodCgroupName {
 				cgroup = pullArgs.sandboxCgroup
 			} else {
 				cgroup = s.config.SeparatePullCgroup
